=begin
Ruby-IXP, Copyright 2006 Stephan Maka
Copyright 2006 Suraj N. Kurapati

This file is part of Ruby-IXP.

Ruby-IXP is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Ruby-IXP is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Ruby-IXP; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA)
=end
require 'socket'
require 'thread'

require 'ixp/fcall'

module IXP
  class IXPException < RuntimeError
  end

  # Encapsulates communication with the IXP server.
  class Connection
    TAG_LIMIT = 2 ** 16

    # Creates a new IXP connection on the given I/O stream.
    def initialize aStream
      @sock = aStream
      @tag = 0

      @orders = {} # ticket => patron
      @service_lock = Mutex.new

      Thread.new {serve_orders}
    end

    # Issues the given request and returns its response.
    def query aReq
      # order coffee
        ticket = @service_lock.synchronize do
          @tag = @tag.next % TAG_LIMIT
        end

        puts "#{Thread.current} placing order #{ticket}" if $DEBUG

        order = @service_lock.synchronize do
          @orders[ticket] = Queue.new
        end

        aReq.fields[:tag] = ticket
        issue_order aReq

      # wait for coffee
        puts "#{Thread.current} waiting for order #{ticket}" if $DEBUG
        coffee = order.deq

      # enjoy coffee
        puts "order #{ticket} is served to #{Thread.current}" if $DEBUG

        if coffee.kind_of? Rerror
          raise IXPException.new(coffee.fields[:ename])
        else
          coffee
        end
    end


    private

    # Waiter issues the given order to the chef, who will make the order.
    def issue_order aOrder
      puts "chef is making order #{aOrder.fields[:tag]}: #{aOrder.inspect}\n" if $DEBUG

      buf = aOrder.to_b
      @sock.write [buf.size + 4].pack('I') + buf
    end

    # Waiter takes finished orders from the chef and serves them to patrons.
    def serve_orders
      while buf = @sock.read(4)
        len, = buf.unpack('I')
        buf = @sock.read(len - 4)

        coffee = Fcall::from_b(buf)
        ticket = coffee.fields[:tag]
        puts "chef finished order #{ticket}: #{coffee.inspect}" if $DEBUG

        order = @service_lock.synchronize do
          @orders.delete ticket
        end

        puts "waking up patron for order #{ticket}" if $DEBUG
        order.enq coffee
      end

      raise "chef has quit!"
    end
  end

  # An IXP client capable of communicating with an IXP server.
  class Client < Connection
    # modes
    OREAD = 0x00
    OWRITE = 0x01
    ORDWR = 0x02
    OEXEC = 0x03
    OEXCL = 0x04
    OTRUNC = 0x10
    OREXEC = 0x20
    ORCLOSE = 0x40
    OAPPEND = 0x80

    # permissions
    DMWRITE = 0x80

    # qid_type
    QTDIR = 0x80

    # Connects this client to an IXP server which operates via a UNIX domain socket at the given path.
    def initialize aSocketPath
      raise 'Nowhere to connect' unless aSocketPath

      super(UNIXSocket.new(aSocketPath.sub(/^unix!/, '')))
      @last_fid = -1

      rv = query(Tversion.new(:version=>'9P2000', :msize=>8192))
      if rv.fields[:version] != '9P2000'
        raise "Unknown 9P version: #{rv.version.inspect}"
      end

      @root_fid = gen_fid
      query(Tattach.new(:fid => @root_fid, :afid => 0, :uname => ENV['USER'], :aname => ''))
    end

    # Opens the given path for I/O operations.
    def open aPath, aMode = 'r'
      if aMode.respond_to? :to_str
        aMode = case aMode.to_str
          when 'r'
            OREAD

          when 'r+'
            ORDWR

          when 'w'
            OTRUNC | OWRITE

          when 'w+'
            OTRUNC | ORDWR

          when 'a'
            OAPPEND | OWRITE

          when 'a+'
            OAPPEND | ORDWR

          else
            raise ArgumentError, "invalid mode string: #{aMode}"
        end
      end

      fid = walk(aPath)

      req = Topen.new(:fid => fid, :mode => aMode)
      resp = query(req)

      type =
        if resp.fields[:qid_type] == QTDIR
          Directory
        else
          File
        end

      io = type.new(self, fid, resp.fields[:iounit])

      if block_given?
        begin
          yield io
        ensure
          io.close
        end
      else
        io
      end
    end

    # Creates a file at the given path with the given permissions.
    def create aPath, aPerms = DMWRITE, aMode = OWRITE
      path = aPath.to_s.squeeze('/').split('/')

      fid = walk(path[0..-2].join('/'))
      query Tcreate.new(
        :fid => fid,
        :name => path.last,
        :perm => aPerms,
        :mode => aMode
      )
      query Tclunk.new(:fid => fid)
    end

    # Deletes the file at the given path.
    def remove aPath
      fid = walk(aPath)
      query Tremove.new(:fid=>fid)
      # No clunking needed
    end

    # Writes the given content to the given path.
    def write aPath, aContent
      open(aPath, 'w') do |f|
        f.write aContent.to_s
      end
    end

    # Reads from the given path and returns the content. If the path is a directory, then all entries in that directory are returned.
    def read aPath
      open(aPath) do |f|
        if f.respond_to? :entries
          f.entries
        else
          f.read
        end
      end
    end

    # Tests if the given path is a file.
    def file? aPath
      open(aPath) {|f| f.instance_of? File} rescue false
    end

    # Tests if the given path is a directory.
    def directory? aPath
      open(aPath) {|f| f.instance_of? Directory} rescue false
    end

    # Tests if the given path exists.
    def exist? aPath
      open(aPath) {true} rescue false
    end


    private

    # Generates and returns a unique file ID.
    def gen_fid
      @last_fid += 1
    end

    # Requests the IXP server for a file ID associated with the given path.
    # result:: [Fixnum] newfid
    def walk aPath
      aPath = aPath.to_s.squeeze('/')

      fid = gen_fid
      query Twalk.new(:fid => @root_fid, :newfid => fid, :nwname => aPath.sub(/^\//, '').split('/'))
      fid
    end
  end

  class File
    def initialize(client, fid, iounit)
      @client = client
      @fid = fid
      @iounit = iounit
      @pos = 0

      @isOpen = true
    end

    # Reads and returns the contents of this file. If there is no content, then *nil* is returned.
    def read
      req = Tread.new(:fid => @fid, :offset => @pos, :count => @iounit)
      resp = @client.query(req)

      if (data = resp.fields[:data]) && (data.size > 0)
        @pos += data.size
        data
      else
        nil
      end
    end

    # Writes the given content to this file and returns the number of bytes that have been written.
    def write aContent
      buf = aContent.to_s

      while buf.size > @iounit
        write buf[0..(@iounit-1)]
        buf = buf[@iounit..-1]
      end

      req = Twrite.new(:fid => @fid, :offset => @pos, :data => buf)
      resp = @client.query(req)

      @pos += resp.fields[:count]
      resp.fields[:count]
    end

    alias << write

    def close
      if @isOpen
        @client.query Tclunk.new(:fid => @fid)
        puts "fid #{@fid} is now closed" if $DEBUG

        @isOpen = false
      else
        puts "fid #{@fid} already closed!" if $DEBUG
      end
    end
  end

  class Directory < File
    def initialize(client, fid, iounit)
      super
      @buf = nil
    end

    # Returns the next entry in this directory.
    def next
      @buf = read unless @buf

      if @buf.size < 2
        nil
      else
        stat = Stat.new(@buf)
        @buf = @buf[(stat.size+2)..-1]
        stat
      end
    end

    # Returns all entries in this directory.
    def entries
      arr = []

      while i = self.next
        arr << i
      end

      arr
    end
  end

  # Parses and holds metadata about directory entries.
  #
  # size[2]
  #     total byte count of the following data
  # type[2]
  #     for kernel use
  # dev[4]
  #     for kernel use
  # qid.type[1]
  #     the type of the file (directory, etc.), represented as a bit vector corresponding to the high 8 bits of the file's mode word.
  # qid.vers[4]
  #     version number for given path
  # qid.path[8]
  #     the file server's unique identification for the file
  # mode[4]
  #     permissions and flags
  # atime[4]
  #     last access time
  # mtime[4]
  #     last modification time
  # length[8]
  #     length of file in bytes
  # name[ s ]
  #     file name; must be / if the file is the root directory of the server
  # uid[ s ]
  #     owner name
  # gid[ s ]
  #     group name
  # muid[ s ]
  #     name of the user who last modified the file
  class Stat
    attr_reader :size, :type, :dev, :qid_type, :qid_version, :qid_path, :mode, :atime, :mtime, :length, :name, :uid, :gid, :muid

    def initialize aBuffer
      buf = aBuffer.to_s

      @size, @type, @dev, @qid_type, @qid_version, @qid_path, @mode, @atime, @mtime, @length = buf.unpack('SSICIa8IIIQ')
      buf = buf[(2+2+4+1+4+8+4+4+4+8)..-1]

      name_size, = buf.unpack('S')
      buf = buf[2..-1]
      @name, = buf.unpack("a#{name_size}")
      buf = buf[name_size..-1]

      uid_size, = buf.unpack('S')
      buf = buf[2..-1]
      @uid, = buf.unpack("a#{uid_size}")
      buf = buf[uid_size..-1]

      gid_size, = buf.unpack('S')
      buf = buf[2..-1]
      @gid, = buf.unpack("a#{gid_size}")
      buf = buf[gid_size..-1]

      muid_size, = buf.unpack('S')
      buf = buf[2..-1]
      @muid, = buf.unpack("a#{muid_size}")
      buf = buf[muid_size..-1]
    end

    def mode_str
      modes = %w(--- --x -w- -wx r-- r-x rw- rwx)

      ((@mode & 0x80000000 != 0) ? 'd' : '-') +
        '-' +
        modes[(mode >> 6) & 7] +
        modes[(mode >> 3) & 7] +
        modes[mode & 7]
    end
  end
end
